'''
Transfer an object to its gear, as an interface.
'''
import asyncio
import datetime

from .AsyncPeriod import AsyncPeriod

from .method_run_when import call_backs
from ensureTaskCanceled import ensureTaskCanceled

gears = {}


class _Gear:
    # last_set_period = {}

    def __init__(self, obj):
        self.obj = obj
        self.periods = {}
        self._unlocked = asyncio.Event()
        self._unlocked.set()
        self.assistant_tasks = []
        self.prev_period = None
        self._current_period: AsyncPeriod = None
        self._period_change_event = asyncio.Event()

    def delete(self):
        '''
        Delete the gear. You'd better delete the gear when it is no more used.

        :return:
        '''
        if self.obj in gears.keys():
            gears.pop(self.obj)
        for task in self.assistant_tasks:
            asyncio.create_task(ensureTaskCanceled(task))

    def _set_intance_gear_callbacks(self, period_names):
        def _set_intance_gear_callbacks(attr, obj, period_names):
            from .run_when import _run_when
            periods2del = []
            for period_name in period_names:  # 新时期被加入，找到可以启动的回调等待
                if period_name in call_backs[attr].keys():  # 启动等待
                    periods2del.append(period_name)
                    for time_method in call_backs[attr][period_name].keys():
                        if asyncio.iscoroutinefunction(attr):
                            @_run_when(obj, time_method, period_name, call_backs[attr][period_name][time_method])
                            async def wrapper():
                                return await asyncio.create_task(attr(obj))
                        else:
                            @_run_when(obj, time_method, period_name, call_backs[attr][period_name][time_method])
                            def wrapper():
                                return attr(obj)

            # 删除该period记录
            [call_backs[attr].pop(period) for period in periods2del]
            # 如果attr没有还要关联的period，则删除该attr的记录
            if not bool(call_backs[attr]):
                call_backs.pop(attr)

        for attr in list(getattr(type(self.obj), '__dict__', {}).values()) + \
                    list(getattr(self.obj, '__dict__', {}).values()):  # 遍历绑定对象的命名空间及其类命名空间，找到实例回调方法对应的类函数
            attr = getattr(attr, '__func__', attr) if type(attr) is classmethod else attr
            if attr in list(call_backs.keys()):
                _set_intance_gear_callbacks(attr, self.obj, period_names)

    def add_periods(self, *new_period_names: str):
        '''
        Dynamically add periods for some object. The first added would be the default.

        :return:
        '''
        for new_period_name in new_period_names:
            if new_period_name in self.periods.keys():
                raise KeyError(f'Period {new_period_name} has already been added.')
            self.periods[new_period_name] = AsyncPeriod(new_period_name, self.obj, self)
            if len(self.periods.keys()) == 1:
                self._set_period(new_period_name)

        self._set_intance_gear_callbacks(new_period_names)

    def get_present_period(self):
        '''
        Get the present period of the target object.

        :return:
        '''
        if self._current_period is not None:
            return self._current_period._name

    def current_set_datetime(self) -> datetime.datetime:
        '''
        Get the UTC datetime when the present period is set.

        :return:
        '''
        if self._current_period is not None:
            return self._current_period._ensured_time

    def get_period_names(self):
        '''
        Get the periods of the target object.

        :return:
        '''
        return tuple(self.periods.keys())

    def sync_set_period(self, period_name: str, slot_num: int = 1):
        '''
        Synchronous version of set_period.
        Set obj to period period_name when unlocked, otherwise PermissionError is raised.

        :param period_name:
        :param slot_num: Attention! Do not use it if you do not understand the parameter!
                slot_num means that only after slot_num times Gear(obj).set_period(period_name,slot_num) run,
                the period of Gear(obj) could really be set to period_name, which is interrupted
                if among these times set_period run, the same period_name with a different slot_num is given.
                Then the procedure is refreshed, the count would be reset.
        :return:
        '''
        return self._set_period(period_name, slot_num)

    def _set_period(self, period_name: str, slot_num: int = 1):
        async def flash_period_change_event():
            self._period_change_event.set()  # todo 类用例 文档
            await asyncio.sleep(0)
            self._period_change_event.clear()

        asyncio.create_task(flash_period_change_event())#todo 暂时没有事件循环的测试

        p = self.periods[period_name]
        p.slots_num_for_true = slot_num
        p.filled_slots_num += 1

    async def set_period(self, period_name: str, slot_num: int = 1):
        '''
        Set obj to period period_name when unlocked, otherwise PermissionError is raised.

        :param period_name:
        :param slot_num: Attention! Do not use it if you do not understand the parameter!
                slot_num means that only after slot_num times Gear(obj).set_period(period_name,slot_num) run,
                the period of Gear(obj) could really be set to period_name, which is interrupted
                if among these times set_period run, the same period_name with a different slot_num is given.
                Then the procedure is refreshed, the count would be reset.
        :return:
        '''

        if self.get_present_period() != period_name:
            await asyncio.create_task(self.wait_outside_period(period_name))
        else:
            await asyncio.create_task(self.wait_inside_period(period_name))
        try:
            if self._unlocked.is_set():
                self._set_period(period_name, slot_num)
            else:
                raise PermissionError('The gear is locked.')
        finally:
            # self.last_set_period[self.obj] = period_name
            if self.get_present_period() != period_name:
                await asyncio.create_task(self.wait_outside_period(period_name))
            else:
                await asyncio.create_task(self.wait_inside_period(period_name))

    async def wait_change_period(self):
        '''
        Wait for the instance when the period is changed.

        :return:
        '''
        await self._period_change_event.wait()

    async def wait_inside_period(self, period_name: str):
        '''
        Wait the time slot when the gear is inside period period_name. As logically, as long as the gear is inside period
        period_name, this coroutine is awaited immediately.

        :param period_name:
        :return:
        '''
        period = self.periods[period_name]
        await asyncio.create_task(period.wait_true())

    async def wait_outside_period(self, period_name: str):
        '''
        Wait the time slot when the gear is outside period period_name. As logically, as long as the gear is outside period
        period_name, this coroutine is awaited immediately.

        :param period_name:
        :return:
        '''
        period = self.periods[period_name]
        await asyncio.create_task(period.wait_false())

    async def wait_enter_period(self, period_name: str):
        '''
        Wait the instant when the gear enters period period_name.

        :param period_name:
        :return:
        '''
        period = self.periods[period_name]
        await asyncio.create_task(period.wait_change_into_true())

    async def wait_exit_period(self, period_name: str):
        '''
        Wait the instant when the gear exits period period_name.

        :param period_name:
        :return:
        '''
        period = self.periods[period_name]
        await asyncio.create_task(period.wait_change_into_false())

    def lock(self):
        '''
        Lock the period of the gear.

        :return:
        '''
        self._unlocked.clear()

    def unlock(self):
        '''
        Unlock the period of the gear after the gear is locked.

        :return:
        '''
        self._unlocked.set()

    async def wait_unlock(self):
        '''
        wait the gear to be unlocked.

        :return:
        '''
        await asyncio.create_task(self._unlocked.wait())
    # def when_enter(self, period_name: str, queue_blocking='abandon'):
    #     return run_when_enter(self.obj, period_name, queue_blocking)
    #
    # def when_exit(self, period_name: str, queue_blocking='abandon'):
    #     return run_when_exit(self.obj, period_name, queue_blocking)
    #
    # def when_inside(self, period_name: str, queue_blocking='abandon'):
    #     return run_when_inside(self.obj, period_name, queue_blocking)
    #
    # def when_outside(self, period_name: str, queue_blocking='abandon'):
    #     return run_when_outside(self.obj, period_name, queue_blocking)


def Gear(obj) -> _Gear:
    if obj not in gears.keys():
        gears[obj] = _Gear(obj)
    return gears[obj]



import asyncio


def _run_when(obj, time_method: str, period_name: str, queue_blocking='abandon'):
    '''
    Decorator, run the decorated when obj is at the exact moment or period.

    :param time_method: enter, exit, inside, outside
    :param obj:
    :param period_name:
    :param queue_blocking: When the decorated is activated too frequently, non_block means run immediately anyway; queue means
                             waits the previous one completing then run the new activated; abandon means abandon the new
                             activated if the previous one has not completes yet.
    :return:
    '''

    def decrator(decorated):
        async def runner():
            if asyncio.iscoroutinefunction(decorated):
                if queue_blocking == 'abandon':
                    pending_task: asyncio.Task = None
                elif queue_blocking == 'queue':
                    q = asyncio.Queue()

                    # Get q item to run decorated
                    async def queue2run_coroutine():
                        while True:
                            await asyncio.create_task(q.get())
                            await asyncio.create_task(decorated())
                            q.task_done()

                    asyncio.create_task(queue2run_coroutine())
            while True:
                # wait the exact time
                from .Gear import Gear
                await asyncio.create_task(getattr(Gear(obj),
                                                  {'enter': 'wait_enter_period',
                                                   'exit': 'wait_exit_period',
                                                   'inside': 'wait_inside_period',
                                                   'outside': 'wait_outside_period'}[time_method])(period_name))
                if not asyncio.iscoroutinefunction(decorated):
                    decorated()
                else:
                    if queue_blocking == 'non_block':
                        asyncio.create_task(decorated())
                    elif queue_blocking == 'abandon':
                        if not bool(pending_task) or pending_task.done():  # previous completes
                            pending_task = asyncio.create_task(decorated())
                        elif bool(pending_task):
                            await pending_task
                    elif queue_blocking == 'queue':
                        q.put_nowait(None)

        runner_task = asyncio.create_task(runner())
        from .Gear import Gear
        Gear(obj).assistant_tasks.append(runner_task)

        return decorated

    return decrator


def run_when_enter(obj, period_name: str, queue_blocking='abandon'):
    '''
    Decorator, run the decorated when obj enters the period.

    :param obj:
    :param period_name:
    :param queue_blocking: When the decorated is activated too frequently, 'non_block' means run immediately anyway; 'queue' means
                             waits the previous one completing then run the new activated; 'abandon' means abandon the new
                             activated if the previous one has not completed yet.
    :return:
    '''
    return _run_when(obj, 'enter', period_name, queue_blocking)


def run_when_exit(obj, period_name: str, queue_blocking='abandon'):
    '''
    Decorator, run the decorated when obj exits the period.

    :param obj:
    :param period_name:
    :param queue_blocking: When the decorated is activated too frequently, 'non_block' means run immediately anyway; 'queue' means
                             waits the previous one completing then run the new activated; 'abandon' means abandon the new
                             activated if the previous one has not completed yet.
    :return:
    '''
    return _run_when(obj, 'exit', period_name, queue_blocking)


def run_when_inside(obj, period_name: str):
    '''
    Decorator, run the decorated when obj is inside the period. The queue blocking style is 'abandon', which means abandon the new
    activated if the previous one has not completed yet.

    :param obj:
    :param period_name:
    :return:
    '''
    return _run_when(obj, 'inside', period_name, 'abandon')


def run_when_outside(obj, period_name: str):
    '''
    Decorator, run the decorated when obj is outside the period. The queue blocking style is 'abandon', which means abandon the new
    activated if the previous one has not completed yet.

    :param obj:
    :param period_name:
    :return:
    '''
    return _run_when(obj, 'outside', period_name, 'abandon')



call_backs = {}


def _when_ins(time_method: str, period_name: str, queue_blocking='abandon'):
    '''
    remenber to instantiate in a coroutine

    :param period_name:
    :param queue_blocking: When the decorated is activated too frequently, 'non_block' means run immediately anyway; 'queue' means
                             waits the previous one completing then run the new activated; 'abandon' means abandon the new
                             activated if the previous one has not completed yet.
    :return:
    '''

    def decorator(f):
        call_backs[getattr(f, '__func__', f) if type(f) is classmethod else f] = \
            call_backs.get(getattr(f, '__func__', f) if type(f) is classmethod else f, {})
        call_backs[getattr(f, '__func__', f) if type(f) is classmethod else f][period_name] = \
            call_backs[getattr(f, '__func__', f) if type(f) is classmethod else f].get(period_name, {})
        call_backs[getattr(f, '__func__', f) if type(f) is classmethod else f][period_name][time_method] = \
            queue_blocking
        return f

    return decorator


def when_enter(period_name: str, queue_blocking='abandon'):
    '''
    Decorate a class method or its instance method, to be run when the class or the instance gear enters period
    period_name. You can set the gear periods later. Remember to use in a coroutine.

    :param period_name:
    :param queue_blocking: When the decorated is activated too frequently, 'non_block' means run immediately anyway; 'queue' means
                             waits the previous one completing then run the new activated; 'abandon' means abandon the new
                             activated if the previous one has not completed yet.
    :return:
    '''
    return _when_ins('enter', period_name, queue_blocking)


def when_exit(period_name: str, queue_blocking='abandon'):
    '''
    Decorate a class method or its instance method, to be run when the class or the instance gear exits period
    period_name. You can set the gear periods later. Remember to use in a coroutine.

    :param period_name:
    :param queue_blocking: When the decorated is activated too frequently, 'non_block' means run immediately anyway; 'queue' means
                             waits the previous one completing then run the new activated; 'abandon' means abandon the new
                             activated if the previous one has not completed yet.
    :return:
    '''
    return _when_ins('exit', period_name, queue_blocking)


def when_inside(period_name: str, ):
    '''
    Decorate a class method or its instance method, to be run when the class or the instance gear is inside period
    period_name. You can set the gear periods later. Remember to use in a coroutine.

    :param period_name:
    :return:
    '''
    return _when_ins('inside', period_name, 'abandon')


def when_outside(period_name: str, ):
    '''
    Decorate a class method or its instance method, to be run when the class or the instance gear is outside period
    period_name. You can set the gear periods later. Remember to use in a coroutine.

    :param period_name:
    :return:
    '''
    return _when_ins('outside', period_name, 'abandon')



import asyncio
import datetime

# todo 以后可以想想实现子集、并集、交集等，最小元素考虑是互斥时期
from loguru import logger


class AsyncPeriod:
    def __init__(self, name, obj, gear):
        self._true_event = asyncio.Event()
        self._false_event = asyncio.Event()
        self._ensure_state(False)
        self._name = name
        self.obj = obj
        self.gear = gear

        self._slots_num_for_true = 1
        self._filled_slots_num = 0
        self._ensured_time = None

    @property
    def slots_num_for_true(self):
        return self._slots_num_for_true

    @slots_num_for_true.setter
    def slots_num_for_true(self, x: int):
        if self._slots_num_for_true != x:  # 更新
            self.filled_slots_num = 0

        self._slots_num_for_true = x

    @property
    def filled_slots_num(self):
        return self._filled_slots_num

    @filled_slots_num.setter
    def filled_slots_num(self, x: int):
        self._filled_slots_num = x
        # 触发
        if self.filled_slots_num >= self.slots_num_for_true:
            self.gear.prev_period = self.gear.get_present_period()
            for period in self.gear.periods.values():
                period._ensure_state(period is self)
            self.gear._current_period = self
            logger.debug(f'set {repr(self.obj)} to period {self._name}.')

            self.filled_slots_num = 0

    def _ensure_state(self, state: bool):
        if state:
            if not self._true_event.is_set():
                self._true_event.set()
                self._ensured_time = datetime.datetime.utcnow()
            if self._false_event.is_set():
                self._false_event.clear()

        else:
            if self._true_event.is_set():
                self._true_event.clear()
            if not self._false_event.is_set():
                self._false_event.set()
                self._ensured_time = None

    def get_state(self):
        return self._true_event.is_set() and not self._false_event.is_set()

    async def wait_true(self):
        await asyncio.create_task(self._true_event.wait())

    async def wait_false(self):
        await asyncio.create_task(self._false_event.wait())

    async def wait_change_into_true(self):
        if self.get_state():
            await asyncio.create_task(self.wait_false())
            await asyncio.create_task(self.wait_true())
        else:
            await asyncio.create_task(self.wait_true())

    async def wait_change_into_false(self):
        if self.get_state():
            await asyncio.create_task(self.wait_false())
        else:
            await asyncio.create_task(self.wait_true())
            await asyncio.create_task(self.wait_false())

    # obj_has_async_exclusive_periods = {}
    #
    # @classmethod
    # def create_obj_periods(cls, obj, *period_names: str):
    #     '''
    #     Initially create periods for some object.
    #
    #     :param obj:
    #     :param period_names: Period names.The first one would be the initial period.
    #     :return:
    #     '''
    #     if obj not in cls.obj_has_async_exclusive_periods.keys():
    #         cls.obj_has_async_exclusive_periods[obj] = cls.obj_has_async_exclusive_periods.get(obj, {})
    #         for period_name in period_names:
    #             cls.obj_has_async_exclusive_periods[obj][period_name] = AsyncGear(period_name, obj)
    #         cls._set_obj_period(obj, period_names[0])
    #     else:
    #         raise KeyError(f'{repr(obj)} has already got some periods! Please use add_period.')
    #
    # @classmethod
    # def add_period(cls, obj, new_period_name: str):
    #     '''
    #     Dynamically add a period for some object.
    #
    #     :return:
    #     '''
    #     if obj not in cls.obj_has_async_exclusive_periods.keys():
    #         cls.create_obj_periods(obj, new_period_name)
    #     else:
    #         cls.obj_has_async_exclusive_periods[obj][new_period_name] = AsyncGear(new_period_name, obj)
    #
    # @classmethod
    # def _get_obj_period(cls, obj, period_name: str):
    #     if obj in cls.obj_has_async_exclusive_periods.keys() and \
    #             period_name in cls.obj_has_async_exclusive_periods[obj].keys():
    #         return cls.obj_has_async_exclusive_periods[obj][period_name]
    #     else:
    #         raise KeyError(f'You did not create {period_name} for {repr(obj)}!')
    #
    # @classmethod
    # def get_obj_present_period(cls, obj):
    #     for name, period in cls._get_obj_periods(obj).items():
    #         if period._get_state():
    #             return name
    #
    # @classmethod
    # def get_obj_period_names(cls, obj):
    #     if obj in cls.obj_has_async_exclusive_periods.keys():
    #         return cls.obj_has_async_exclusive_periods[obj].keys()
    #     else:
    #         raise KeyError(f'You did not create any AsyncGear for {repr(obj)}!')
    #
    # @classmethod
    # def _get_obj_periods(cls, obj):
    #     if obj in cls.obj_has_async_exclusive_periods.keys():
    #         return cls.obj_has_async_exclusive_periods[obj]
    #
    # @classmethod
    # def _set_obj_period(cls, obj, period_name: str):
    #     if cls.get_obj_present_period(obj) != period_name:
    #         for name, period in cls._get_obj_periods(obj).items():
    #             # 目标
    #             if name == period_name:
    #                 period._ensure_state(True)
    #             else:
    #                 period._ensure_state(False)
    #             logger.debug(f'set {repr(obj)} to period {period_name}.')
    #
    # @classmethod
    # async def set_obj_period(cls, obj, period_name: str, slot_num: int = 1):
    #     '''
    #     Set obj to period period_name.
    #
    #     :param obj:
    #     :param period_name:
    #     :param slot_num: Attention! Do not use it if you do not understand it!
    #             slot_num means that only after slot_num times AsyncGear.set_obj_period(obj,period_name,slot_num) run
    #             (present time included), the period of obj gear could really be set to period_name, which is interrupted
    #             if among these times set_obj_period run a different slot_num is given. Then the procedure is refreshed.
    #     :return:
    #     '''
    #     p = cls._get_obj_period(obj, period_name)
    #     p.slots_num_for_true = slot_num
    #     p.filled_slots_num += 1
    #
    #     # cls._set_obj_period(obj, period_name)
    #     # if cls.get_obj_present_period(obj) != period_name:
    #     #     await asyncio.create_task(AsyncGear.wait_outside_period(obj, period_name))
    #     # else:
    #     #     await asyncio.create_task(AsyncGear.wait_inside_period(obj, period_name))
    #     # await asyncio.create_task(AsyncGear.wait_inside_period(obj, period_name))
    #
    # @classmethod
    # async def wait_inside_period(cls, obj, period_name: str):
    #     period: cls = cls._get_obj_period(obj, period_name)
    #     await asyncio.create_task(period._wait_true())
    #
    # @classmethod
    # async def wait_outside_period(cls, obj, period_name: str):
    #     period: cls = cls._get_obj_period(obj, period_name)
    #     await asyncio.create_task(period.wait_false())
    #
    # @classmethod
    # async def wait_enter_period(cls, obj, period_name: str):
    #     period: cls = cls._get_obj_period(obj, period_name)
    #     await asyncio.create_task(period.wait_change_into_true())
    #
    # @classmethod
    # async def wait_exit_period(cls, obj, period_name: str):
    #     period: cls = cls._get_obj_period(obj, period_name)
    #     await asyncio.create_task(period._wait_change_into_false())



from .Gear import Gear
from .run_when import run_when_inside, run_when_exit, run_when_enter, run_when_outside
from .method_run_when import when_enter, when_exit, when_inside, when_outside

__all__ = [Gear, run_when_inside, run_when_exit, run_when_enter, run_when_outside, when_enter, when_exit,
           when_inside, when_outside]
